angular.module('beamng.apps')
.directive('driftSensor', ['bngApi', 'StreamsManager', 'UiUnits', '$interval', function (bngApi, StreamsManager, UiUnits, $interval) {
  return {
    template: '<canvas width="300" height="165"></canvas>',
    replace: true,
    restrict: 'EA',
    link: function (scope, element, attrs) {
      var streamsList = ['sensors','electrics','wheelInfo','stats'];
      StreamsManager.add(streamsList);
    
      scope.$on('$destroy', function () {
        StreamsManager.remove(streamsList);
      });

      var c = element[0]
        , appElement = c.parentElement.parentElement// it's buried two deep apparently :s
        , appLeft = appElement.offsetLeft
        , appTop = appElement.offsetTop
        , ctx = c.getContext('2d')
        , slX = 0
		, slY = 0
		, sdX = 0
		, sdY = 0
		, UITheme = 'dark'
		, EX = 0
		, EY = 0
		, xScale = c.width/200 // scaledPos * scale(x or y) + Transformation; 
		, yScale = c.height/110
		, xTransform = c.width/2
		, yTransform = c.height/1.1
      ;		
	  
	  var posHist = [[0,0,false]] // [x,y bDrifting]
		, posDiffHist = []
		, posDiff = []
		, airspeed = 0
		, travelAngleDegs = 0
		, yawDegs = 0
		, slipAngle = 0
		, slipAngleMag = 0
		, lastSlipAngleMag = 0
		, slipAngleMagDiffHist = []
		, slipAngRad = 0
		, absSlipAngleMag = 0
		, overallSlip = 0
		, gX2abs = 0
		, lastG = 0
		, newG = 0
		, GDiffHist = []
	  ;
		
	  var bDrifting = false
		, gThreshold = 3 // 1/3 G to initiate a drift
		, gCrashThreshold = 15 // initial crash threshold. Changes with speed.
		, bCrashed = false
		, angleCrashThreshold = 15 // threshold for change a crash can make to your facing.
		, minDriftSpeed = 1
		, minDriftAngle = 5 // minimum angle to be considered a drift.
		, driftSide = 0
		, driftTransfers = 0
		, mostDriftTransfers = 0
		, driftTransfersHist = []
		, driftPoints = 0
		, driftPointsHist = [0,0]
		, mostDriftPoints = 0
		, driftPointsTotal = 0
		, eventMultiplier = 0
		, multiplier = 0
		, tappingBonus = 0
		, multiplierColour = 0
		, DriftingIndiColour = 'Black'
	  ;

	  // remeber to change the app:resize if you want it all to work
	  //ctx.setTransform(c.width/200, 0, 0, c.height/200, c.width/2,c.height/2);
	  ctx.setTransform(xScale, 0, 0, yScale, xTransform,yTransform);
	  
	  var fontStyle = 'normal'
		, fontVariant = 'normal'
		, fontWeight = 'bold'
		, fontSize = 10
		, fontFamily = '"Lucida Console"'
		, pointsBox = []
		, totalBox = []
		, lastPointsBox = []
		, mostPointsBox = []
		, multiplierBox = []
		, toggleBox = []
	  ;
	  
	  function setFontSize(context2d,sizeNum){
		fontSize = sizeNum;
		compileFont(context2d);
	  }
	  
	  function setFontWeight(context2d,weight){
		fontWeight = weight;
		compileFont(context2d);
	  }
	  
	  function compileFont(context2d){	
		context2d.font = 
			fontStyle + ' ' + 
			fontVariant + ' ' + 
			fontWeight + ' ' + 
			fontSize + 'px' + ' ' + 
			fontFamily
		; 
	  }
  
		function boxForText(_text,_x,_y,_padding){ 
			var _boxLength = (fontSize / 2) * _text.length + 2 + (_padding * 2); // (half the fontSize) * boxLength;
			var _boxHeight = fontSize + (_padding * 2);
			var _boxX = _x - (_boxLength / 2); // assume center
			if(ctx.textAlign == 'left') {_boxX = _x - fontSize};
			if(ctx.textAlign == 'right') {_boxX = _x - (_boxLength -(fontSize / 2))};
			var _boxY = _y - (fontSize / 2); // assume middle
			// TODO add other Y calculations
			
			var _centerX = _boxX + (_boxLength / 2);
			var _centerY = _boxY + (_boxHeight / 2);
			
			var _array = [_boxX,_boxY,_boxLength,_boxHeight,_centerX,_centerY,fontSize];
			return _array;
		}
		
		function renderBox(_boxArray,_bStroke,_bFill){
			if(_bFill){
				ctx.fillRect(_boxArray[0], _boxArray[1], _boxArray[2], _boxArray[3]);
			}			
			if(_bStroke){
				ctx.strokeRect(_boxArray[0], _boxArray[1], _boxArray[2], _boxArray[3]);
			}
		}
	  
	  // To sort the wheels from their messy non alphabetised list.
	  // To sort the wheels from their messy non alphabetised list.
      var sortedWheelIndices = null;

      var _sortedWheelIndicesFunc = function (wheelInfo){ // TODO: duplicated in WeightDistribution app
        function getWheelNames(wheelInfo){
          var names=[];
          for(var i in wheelInfo)
            names.push(wheelInfo[i][0]);
          return names;
        }

        function wheelNamesToIndices(wheelInfo, wheelNames){
          var indices=[];
          for(var i in wheelNames)
            for(var j in wheelInfo)
              if (wheelNames[i] == wheelInfo[j][0])
                indices.push(j);
          return indices;
        }
        return wheelNamesToIndices(wheelInfo, getWheelNames(wheelInfo).sort()); // alphabetical sort suits our needs: FL < FR < RL < RR
      };
	  
	  function numberColour(num) {		
		var _colour = ''; 		
		if(num < -0.99)
			_colour = 'Red';
		else if (num > 0.99)
			_colour = 'Green';
		else
			_colour= 'white'; // inside the tolerances			
		return _colour;
	  };

	  function resetTotal() {
		driftPointsHist = [];
		driftPointsTotal = 0;
	  }
	  
	  function clearCurrentPoints() {
		driftPoints = 0;
		driftTransfers = 0;
		tappingBonus = 0;
		eventMultiplier = 0;
	  }
	  
	  function addPoints() {
		if(bDrifting) {
			driftPoints = driftPoints + (0.1 * multiplier);
			//driftPoints = driftPoints++;
		} else if(!bDrifting && ((driftPoints > 0) || (eventMultiplier > 0))) {
			if((driftPoints > 5) && (airspeed > minDriftSpeed) && !bCrashed){
				driftPointsHist.unshift(driftPoints * eventMultiplier);
				if(driftPointsHist.length >= 100){
					driftPointsHist.pop(); // removes the last element
				}
				mostDriftPoints = Math.max.apply(null,driftPointsHist);
				driftPointsTotal += driftPointsHist[0]; // add last points to total
				// check if we hit a new record for drift transfers
				driftTransfersHist.unshift(driftTransfers);
				mostDriftTransfers = Math.max.apply(null,driftTransfersHist);
			}
			clearCurrentPoints();
		} else {
			// nothing
		}
		
	  }
	  
	  function bSlipping(wheel){				
		if(wheel[5] > 2)
			return true;
		else
			return false;
	  };
	  
	  function getAngle(_opp,_adj){
		var _angle = Math.atan2(_opp,_adj);
		
		return _angle;
	  }
	  

	 
	  var driftDowntime = 0;
	  var downtimeCheckInt = 0.1; //secs
	  var gripping = true;
	  function checkDriftDowntime() {
		if(gripping) {
			driftDowntime += downtimeCheckInt;
		} else {
			driftDowntime = 0;
		}
	  }
	  
	  var lastDeformedBeams = 0;
	  function updateDeformedBeams(newDeformedBeams) {
		lastDeformedBeams = newDeformedBeams;
	  }
	 
	  var flashTick = 0; // used for the flashing parts of the display
	  function flashTickFunc(){
		flashTick = Math.abs(flashTick - 1);
	  }	 

	  function toggleTheme() {
		if(UITheme == 'light'){
			UITheme = 'dark';
		}else{
			UITheme = 'light';
		};
	  }	  
	  
	  // running loops
	  setInterval(addPoints,100);
	  setInterval(checkDriftDowntime,downtimeCheckInt * 1000);
	  setInterval(flashTickFunc,333);
	  
      scope.$on('streamsUpdate', function (event, data) {
	  
		// sort the wheels
		var wheels = data.wheelInfo;

		if (!sortedWheelIndices || (wheels && wheels.length != sortedWheelIndices.length))
		  sortedWheelIndices = _sortedWheelIndicesFunc(wheels);
		  
		var FL = wheels[sortedWheelIndices[0]];
		var FR = wheels[sortedWheelIndices[1]];
		var RL = wheels[sortedWheelIndices[2]];
		var RR = wheels[sortedWheelIndices[3]];

		var wheelCount = 0;
		/* wheels format:
		0  wd.name
		1  wd.radius
		2  wd.wheelDir
		3  w.angularVelocity
		4  w.lastTorque
		5  drivetrain.wheelInfo[wd.wheelID].lastSlip
		6  wd.lastTorqueMode
		7  brake pressure
		8  a constant of some kind?
		*/
		
		// Glean angle of travel from position whcih is taken and compared from moment to moment 	
		airspeed = data.electrics.airspeed;
		
		// store history of positions
		posHist.unshift([data.sensors.position.x,data.sensors.position.y,bDrifting]);
		if(posHist.length >= 1000){ posHist.pop() };
		
		posDiff = [posHist[0][0] - posHist[1][0],posHist[0][1] - posHist[1][1]];
		
		// store history of position differences
		posDiffHist.unshift(posDiff);
		if(posDiffHist.length >= 100){ posDiffHist.pop() };
		

		travelAngleDegs = getAngle(posDiff[0],posDiff[1]) * 180 / Math.PI;
		if(travelAngleDegs < 0) {travelAngleDegs += 360;}
		yawDegs = Math.abs(data.sensors.yaw - Math.PI) * 180 / Math.PI;
		slipAngle = travelAngleDegs - yawDegs;
		if(airspeed < minDriftSpeed){slipAngle = 0;}
		if(slipAngle < 0){slipAngle += 360;}
		
		//move old data into reference and replace with new
		// old
		lastSlipAngleMag = slipAngleMag; 
		// new
		if(slipAngle > 180){
			slipAngleMag = slipAngle - 360;
		} else {
			slipAngleMag = slipAngle;
		} 
		slipAngleMagDiffHist.unshift(slipAngleMag - lastSlipAngleMag);
		if(slipAngleMagDiffHist.length >= 10){
			slipAngleMagDiffHist.pop(); // removes the last element
		}
		
		absSlipAngleMag = Math.round(Math.abs(slipAngleMag));
		

		

		
		lastG = newG;
		newG = Math.sqrt((data.sensors.gx2 * data.sensors.gx2) * (data.sensors.gy2 * data.sensors.gy2));
		GDiffHist.unshift(newG - lastG);
		if(GDiffHist.length >= 10){
			GDiffHist.pop(); // removes the last element
		}	
				
		gX2abs = Math.abs(data.sensors.gx2); // Gs being pull along the X axis of the car (relative)

		
		// Drifting starts if atleast one of the two rear wheels are slipping and the Gs in the x axis are significant
		if( (bSlipping(RL) || bSlipping(RR)) && (absSlipAngleMag > minDriftAngle) && (gX2abs > gThreshold) && (!bCrashed)) {
			bDrifting = true;
			gripping = false;				
		}
		// Drifting then stops if both wheels aren't slipping or you are going less than 1mph.
		if(  !( bSlipping(RL) || bSlipping(RR) ) || (airspeed < minDriftSpeed) ){
			gripping = true;
			if(driftDowntime > 0.2) {
				bDrifting = false;
			}
		}
		
		overallSlip = (RL[5] + RR[5]) / 2;

		// if we hit something fairly hard the drift is not counted
		if(data.stats.beams_deformed != lastDeformedBeams) {
			for(var i = 0; i < GDiffHist.length; i++){
				if((Math.max.apply(null,GDiffHist) > gCrashThreshold) || (Math.min.apply(null,GDiffHist) < -gCrashThreshold))
				{
					bDrifting = false;
					bCrashed = true;
					setTimeout(function(){bCrashed = false},2000);
					break;
				}
			}
			
			if(!bCrashed && bDrifting){
				tappingBonus++; 
				// because you've just tapped something without totaling the car or losing the drift.
			}

			updateDeformedBeams(data.stats.beams_deformed);
		}
		
		// calculate a multiplier based on current speed and slip angle. the faster the drift with more angle, the better
		var speedMultiplier = airspeed / 4; // every 4 mph increases the mulitplier by 1
		var rubberMultiplier = overallSlip / 4; 
/*		if(speedMultiplier < 1){
			speedMultiplier = 1;
		}
*/		gCrashThreshold = (9.81 / 4) * airspeed; // 1/2 G per 10mph because the GForce info is in newtons?
		// Logic for drift transfer multipliers (going from one into another)
		if(bDrifting) { 
			if(driftSide == 0) {
				if(slipAngleMag > 0) {driftSide = 1;} else {driftSide = -1;}
				driftTransfers++;
			}
			
			if((driftSide == 1) && (slipAngleMag < -minDriftAngle)) {
				driftSide = -1;
				driftTransfers++;
			} else if((driftSide == -1) && (slipAngleMag > minDriftAngle)) {
				driftSide = 1;
				driftTransfers++;
			} else {
				// no change in drift side.
			}
		} else {
			driftSide = 0;
			//driftTransfers = 0; //  made 0 when points are allocated
		}	
		
		multiplier = rubberMultiplier + speedMultiplier + Math.abs(slipAngleMag / 10); // add proximity to objects as a modifier!!!
		eventMultiplier = tappingBonus + driftTransfers;
		multiplierColour = Math.round((255 / 14) * multiplier) + 127;
									//(target max / expected max) * variable with the expected max
		
		// setting alpha value based on multiplier to show a good drift by intensity of colour!
		if(bDrifting) { 
			var _greenvalue = 0.2 + ((airspeed / 25) * 0.4) + ((overallSlip / 20) * 0.4); //maxes out at 40mph
			DriftingIndiColour = "RGBA(0,200,0," + _greenvalue.toFixed(3) + ")";
		} else {
			DriftingIndiColour = "RGBA(200,0,0,0.8)";
		}		
		
		

		//UI calculations
		var gaugeRad = 84;
		var gaugeThickness = 14;
		var gaugeOutlineThickness = 1;
		var gaugeInRad = gaugeRad - (gaugeThickness / 2);
		var gaugeOutRad = gaugeRad + (gaugeThickness / 2);
		var gaugeTextColour = "RGBA(255,255,255,1)"; // text on readouts and highlighted gauge values
		var gaugeShadowTextColour = "RGBA(0,0,0,0.7)"; // label text and unhighlighted gauge values
		var gaugeShadowColour = "RGBA(255,255,255,0.5)"; // backdrop shadow of gauge, fill of boxes
		var gaugeShadowOutlineColour = "RGBA(0,0,0,0.8)"; // Line surrounding gauge and boxes
		var crashColour = "RGBA(255,0,0,0.5)"; // flash colour for crashing
		var crashShadowColour = "red"; // flash colour for crashing
		var _slipPerc = (overallSlip / 18);
		var slipRateColour = "blue"; // colour shown for burning rubber
		var shadowBlur = 2; // how dispersed the shadows are
		if(UITheme == 'dark'){
			gaugeTextColour = "RGBA(255,255,255,1)";
			gaugeShadowTextColour = "RGBA(255,255,255,0.5)";
			gaugeShadowColour = "RGBA(0,0,0,0.8)";
			gaugeShadowOutlineColour = "RGBA(0,0,0,0.8)";
		}
		var needleLimit = 94;
		var gaugeFillThickness = gaugeThickness - 2;
		var baseTextSize = gaugeFillThickness - 4;
		var displaySlipAngle = slipAngleMag; // limited to 90deg either way to not break UI look
		if(!bDrifting) {displaySlipAngle = 0}
		if(displaySlipAngle > 90) {displaySlipAngle = 90;}
		if(displaySlipAngle < -90) {displaySlipAngle = -90;}
		
		//position of slip line on end on UI
		slX = Math.sin(displaySlipAngle * Math.PI / 180) * needleLimit;
		slY = Math.cos(displaySlipAngle * Math.PI / 180) * needleLimit;
		//position of slip dot marker on guage
		sdX = Math.sin(displaySlipAngle * Math.PI / 180) * gaugeRad;
		sdY = Math.cos(displaySlipAngle * Math.PI / 180) * gaugeRad;
		
		//Render display
		ctx.shadowBlur = 0;
        setFontSize(ctx,10);
		ctx.textAlign = "center";
        ctx.textBaseline = "middle";

        ctx.clearRect(-100,-100,200,200);
        // ctx.clearRect(0, 0, 300, 300);

        // Circles
        ctx.strokeStyle = "RGBA(0,0,0,1)";
        ctx.lineWidth = 5;
        ctx.fillStyle = "RGBA(0,0,0,0.8)";
	
		//Gauge shadow
			ctx.beginPath();
		if(bDrifting){
			ctx.shadowBlur = shadowBlur;
			ctx.shadowColor = slipRateColour;
		}
		ctx.strokeStyle = gaugeShadowColour;
		ctx.lineCap = 'round';
		ctx.lineJoin = 'round';
		ctx.lineWidth = gaugeThickness;
		ctx.arc(0,0,gaugeRad,0 * Math.PI,1 * Math.PI,1);
		if(bCrashed && flashTick){ctx.shadowBlur = shadowBlur;ctx.shadowColor = crashShadowColour;}
		ctx.stroke();
		
			ctx.beginPath();
		ctx.strokeStyle = "RGBA(0," + (_slipPerc * 100).toFixed(0) + "," + (_slipPerc * 255).toFixed(0) + "," + (0.5 + (_slipPerc * 0.3)).toFixed(0) + ")";
		if(bCrashed && flashTick){ctx.strokeStyle = crashColour; ctx.shadowBlur = shadowBlur;ctx.shadowColor = crashShadowColour;}
		ctx.lineWidth = (gaugeOutlineThickness + (_slipPerc * (1 + gaugeOutlineThickness))).toFixed(0); if(bCrashed && flashTick){ctx.lineWidth = 3;}
		ctx.arc(0,0,gaugeInRad,0 * Math.PI,1 * Math.PI,1);
		ctx.stroke();
			ctx.beginPath();
		ctx.arc(0,0,gaugeOutRad,0 * Math.PI,1 * Math.PI,1);
		ctx.stroke();
		ctx.clearRect(-100,0,200,10);
		ctx.shadowBlur = 0; // reset
		
		//Gauge fill
		ctx.beginPath();
		ctx.strokeStyle = DriftingIndiColour; //"RGBA(255,0,0,0.8)"
		ctx.lineWidth = gaugeFillThickness;
		//ctx.lineCap = 'round';
		var arcDir = 0; if(displaySlipAngle < 0){arcDir = 1};
		ctx.arc(0,0,gaugeRad,1.5 * Math.PI,(displaySlipAngle * Math.PI / 180) + (1.5 * Math.PI),arcDir);
		ctx.moveTo(sdX,-sdY);
		ctx.lineTo(slX,-slY); // negative because of the render of the UI... pixels render left to right/ top to bottom
		ctx.stroke();
		ctx.shadowBlur = 0; // reset
		
		setFontWeight(ctx,'bold');
		ctx.fillStyle = gaugeTextColour;
		ctx.fillText('x' + (Math.abs(displaySlipAngle) / 10).toFixed(0),sdX,-sdY);
		
		// Multiplier markers
		for(var i = 2; i < 9; i = i + 2) {
			//positions of text on guage
			// right side
			var _angle = i * 10;
			setFontWeight(ctx,'normal');
			ctx.fillStyle = gaugeShadowTextColour; 
			ctx.shadowBlur = 0; // reset
			if((displaySlipAngle >= _angle) && flashTick) {
				ctx.shadowBlur = 5;
				ctx.shadowColor = "yellow";
				setFontWeight(ctx,'bold');
				ctx.fillStyle = gaugeTextColour;
			}
			ctx.fillText('x' + i.toFixed(0), (Math.sin(_angle * Math.PI / 180) * gaugeRad), -(Math.cos(_angle * Math.PI / 180) * gaugeRad));
			// left side
			setFontWeight(ctx,'normal');
			ctx.fillStyle = gaugeShadowTextColour;
			ctx.shadowBlur = 0; // reset
			if((displaySlipAngle <= -_angle) && flashTick) {
				ctx.shadowBlur = 5;
				ctx.shadowColor = "yellow";
				setFontWeight(ctx,'bold');
				ctx.fillStyle = gaugeTextColour;
			}
			ctx.fillText('x' + i.toFixed(0), -(Math.sin(_angle * Math.PI / 180) * gaugeRad), -(Math.cos(_angle * Math.PI / 180) * gaugeRad));
		}
		
		ctx.shadowBlur = 0; // reset
/*		
		var baseLine = 80;
		ctx.beginPath();
		ctx.strokeStyle = "RGBA(0,255,255,0.8)";
		ctx.lineWidth = 1;
		ctx.lineCap = 'round';
		ctx.moveTo(-baseLine,baseLine - Math.abs(slipAngleMagDiffHist[0]));
		for(var i = 1; i < slipAngleMagDiffHist.length; i++) {

			ctx.lineTo(-baseLine + (i * 20),baseLine - Math.abs(slipAngleMagDiffHist[i])); // negative because of the render of the UI... pixels render left to right/ top to bottom
			
		}
		ctx.stroke(); // draw line
		
		ctx.beginPath();
		ctx.strokeStyle = "RGBA(0,100,255,0.8)";
		ctx.moveTo(-baseLine,baseLine - angleCrashThreshold);
		ctx.lineTo(-baseLine + (slipAngleMagDiffHist.length * 20),baseLine - angleCrashThreshold); 
		ctx.stroke(); // draw line
		
		//GDiffHist
		
		ctx.beginPath();
		ctx.strokeStyle = "RGBA(255,255,0,0.8)";
		ctx.lineWidth = 1;
		ctx.lineCap = 'round';
		ctx.moveTo(-baseLine,baseLine - Math.abs(GDiffHist[0]));
		for(var i = 1; i < GDiffHist.length; i++) {

			ctx.lineTo(-baseLine + (i * 20),baseLine - Math.abs(GDiffHist[i])); // negative because of the render of the UI... pixels render left to right/ top to bottom
			
		}
		ctx.stroke(); // draw line
		
		ctx.beginPath();
		ctx.strokeStyle = "RGBA(255,100,0,0.8)";
		ctx.moveTo(-baseLine,baseLine - gCrashThreshold);
		ctx.lineTo(-baseLine + (GDiffHist.length * 20),baseLine - gCrashThreshold); 
		ctx.stroke(); // draw line
*/		
		//posHist - drift trail graphic
/*		
		ctx.beginPath();
		var zoom = 2;
		ctx.lineWidth = zoom * zoom;
		ctx.lineCap = 'round';
		ctx.moveTo(0,0);
		var xNormaliser = -(posHist[0][0] * zoom);
		var yNormaliser = -(posHist[0][1] * zoom);
		if(posHist[0][2]){
			ctx.strokeStyle = "RGBA(0,200,0,0.8)"
		}else{
			ctx.strokeStyle = "RGBA(255,0,0,0.8)"
		};		
		for(var i = 1; i < posHist.length; i++) {
			// If there is a difference in the bDrifting state for either position
			if(posHist[i - 1][2] != posHist[i][2]){
				ctx.stroke();
				ctx.beginPath();
				// if current pos says drift
				if(posHist[i][2]){
					// behind us should be stroked with the non drift colour
					ctx.strokeStyle = "RGBA(255,0,0,0.8)";
				}else{ // if not
					// behind us was a drift and should be the drift colour before we continue
					ctx.strokeStyle = "RGBA(0,200,0,0.8)";
				};
				
			}
		
			ctx.lineTo(-((posHist[i][0] * zoom) + xNormaliser),(posHist[i][1] * zoom) + yNormaliser); // negative because of the render of the UI... pixels render left to right/ top to bottom
			
		}
		//ctx.stroke(); // draw line
*/
		
		//center of gauge
			// background
		ctx.fillStyle = gaugeShadowColour;
		setFontWeight(ctx,'bold');
		
		// measure out boxes that use the same font size then render them and move onto the next set
		setFontSize(ctx,(baseTextSize * 2));
		pointsBox = boxForText('Points', 0, -61, 3);
		totalBox = boxForText('Total', 0, -15, 3);
		setFontSize(ctx,(baseTextSize * 1.5));
		lastPointsBox = boxForText('00000', -22, -38, 3);
		mostPointsBox = boxForText('00000', 22, -38, 3);
		setFontSize(ctx,(baseTextSize * 1.3));
		multiplierBox = boxForText('00', 38, -54, 3);
		toggleBox = boxForText('0', 40, -20, 1);
		
		
		ctx.fillStyle = gaugeShadowColour;
		if(bDrifting){
			ctx.shadowBlur = shadowBlur;
			ctx.shadowColor = slipRateColour;
		}
		ctx.strokeStyle = "RGBA(0," + (_slipPerc * 100).toFixed(0) + "," + (_slipPerc * 255).toFixed(0) + "," + (0.5 + (_slipPerc * 0.3)).toFixed(0) + ")";
		if(bCrashed && flashTick){ctx.strokeStyle = crashColour; ctx.shadowBlur = shadowBlur;ctx.shadowColor = crashShadowColour;}
		ctx.lineWidth = (gaugeOutlineThickness + (_slipPerc * (gaugeOutlineThickness))).toFixed(0); if(bCrashed && flashTick){ctx.lineWidth = 3;}
		
		renderBox(pointsBox,1,1);
		renderBox(totalBox,1,1);
		renderBox(lastPointsBox,1,1);
		renderBox(mostPointsBox,1,1);
		//renderBox(multiplierBox,1,1);
		
		ctx.shadowBlur = 0; // reset
		// box labels
		ctx.textAlign = "left";
        ctx.textBaseline = "top";
		setFontSize(ctx,baseTextSize * 0.75);
		ctx.fillStyle = gaugeShadowTextColour;
		ctx.fillText('Points', pointsBox[0], pointsBox[1]);
		ctx.fillText('Total', totalBox[0], totalBox[1]);
		ctx.fillText('Last', lastPointsBox[0], lastPointsBox[1]);
		ctx.fillText('Most', mostPointsBox[0], mostPointsBox[1]);
		ctx.fillText('X', multiplierBox[0], multiplierBox[1]);
		
			// Foreground
		ctx.textAlign = "center";
        ctx.textBaseline = "middle";

		ctx.fillStyle = gaugeTextColour;
		setFontSize(ctx,pointsBox[6]);
		ctx.fillText(driftPoints.toFixed(0), pointsBox[4], pointsBox[5]);
		ctx.fillText(driftPointsTotal.toFixed(0), totalBox[4], totalBox[5]);
		
		setFontSize(ctx,lastPointsBox[6]);
		ctx.fillText(driftPointsHist[0].toFixed(0), lastPointsBox[4], lastPointsBox[5]);
		ctx.fillText(mostDriftPoints.toFixed(0), mostPointsBox[4], mostPointsBox[5]);
	 
		setFontSize(ctx,multiplierBox[6]);
		ctx.fillText(eventMultiplier.toFixed(0), multiplierBox[4], multiplierBox[5]);		
/*		ctx.fillText(overallSlip.toFixed(2), 0, 0);		
		ctx.fillText(appElement.nodeName, 60, -40);		
		ctx.fillText(appLeft + ',' + appTop, 60, -20);		
		ctx.fillText(EX + ',' + EY, 60, 0);	*/	
		
		//toggle buttons
			ctx.beginPath();
		ctx.strokeStyle = "RGBA(255,255,255,0.8)";
		ctx.fillStyle = "white";
		if(UITheme == 'light') {
			ctx.strokeStyle = "RGBA(0,0,0,0.8)";
			ctx.fillStyle = "black";
		}
		ctx.lineWidth = gaugeOutlineThickness;
		ctx.arc(toggleBox[4],toggleBox[5],toggleBox[2] / 2,2 * Math.PI,0,1);
		ctx.fill();
		ctx.stroke();
	 });

	 
	 // Event listeners
	  
	c.addEventListener('click', function(event) {
		EX = event.pageX - appLeft;
        EY = event.pageY - appTop;
		var tBX = (totalBox[0] * xScale) + xTransform; // TotalboxX in pixels from the top left.
		var tBL = totalBox[2] * xScale; // Totalbox length in pixels.
		var tBY = (totalBox[1] * yScale) + yTransform;
		var tBH = totalBox[3] * yScale;
		
		var tgBX = (toggleBox[0] * xScale) + xTransform; // toggleBoxX in pixels from the top left.
		var tgBL = toggleBox[2] * xScale; // toggleBox length in pixels.
		var tgBY = (toggleBox[1] * yScale) + yTransform;
		var tgBH = toggleBox[3] * yScale;
			
		if((EX > tBX) && (EX < (tBX + tBL))){
			if((EY > tBY) && (EY < (tBY + tBH))){
				resetTotal();
			}
		}
		
		if((EX > tgBX) && (EX < (tgBX + tgBL))){
			if((EY > tgBY) && (EY < (tgBY + tgBH))){
				toggleTheme();
			}
		}
		
	});
	  
	  scope.$on('VehicleChange', function (event, data) {
        sortedWheelIndices = null;
      });

      scope.$on('app:resized', function (event, data) {
		c.width = data.width;
        c.height = data.height;
        //ctx.setTransform(c.width/200, 0, 0, c.height/200, c.width/2,c.height/2);
        ctx.setTransform(xScale, 0, 0, yScale, xTransform,yTransform);
	  
		appLeft = appElement.offsetLeft
        appTop = appElement.offsetTop
      });
	 	  
    }
  };
}]);